﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Microsoft.Speech.AudioFormat;
using Microsoft.Speech.Recognition;

namespace Ozeki.Media
{
    public class MSSpeechPlatformSTT : ISpeechToText
    {
        private AudioFormat _audioFormat;
        private SpeechAudioFormatInfo _speechFormat;
        private SpeechRecognitionEngine _recognition;
        private IEnumerable<string> _choices;
        private object _sync;

        public MSSpeechPlatformSTT()
        {
            _sync = new object();
        }

        #region ISpeechToText Members

        public string Name
        {
            get { return "MSSpeechPlatform"; }
        }

        public SpeechRecognizerInfo RecognizerInfo
        {
            get
            {
                var info = new SpeechRecognizerInfo();
                if (_recognition == null)
                    return info;

                return GetRecognizerInfo(_recognition.RecognizerInfo);
            }
        }

        public bool Initialized { get; private set; }

        public event EventHandler<SpeechDetectionEventArgs> WordRecognized;
        public event EventHandler<SpeechDetectionEventArgs> WordHypothesized;
        public event EventHandler SpeechDetected;
        public event EventHandler RecognitionCompleted;

        public bool Init(AudioFormat audioFormat, IEnumerable<string> choices)
        {
            _audioFormat = audioFormat;
            _choices = choices;

            return InitSTT();
        }

        
        public void StartRecognition(Stream stream)
        {
            lock (_sync)
            {
                if (_recognition == null)
                    return;

                _recognition.SetInputToAudioStream(stream, _speechFormat);

                if (stream.Position == stream.Length)
                {
                    OnRecognitionCompleted();
                    return;
                }

                _recognition.RecognizeAsync(RecognizeMode.Multiple);
            }
        }

        public void StopRecognition()
        {
            lock (_sync)
            {
                if (_recognition == null)
                    return;

                _recognition.RecognizeAsyncStop();
            }
        }

        public IEnumerable<SpeechRecognizerInfo> GetRecognizers()
        {
            var list = new List<SpeechRecognizerInfo>();

            var installed = SpeechRecognitionEngine.InstalledRecognizers();
            foreach (var sr in installed)
            {
                var info = GetRecognizerInfo(sr);
                list.Add(info);
            }

            return list;
        }

        public bool ChangeRecognizer(string recognizerID)
        {
            if (!Initialized)
                return false;

            StopRecognition();
            CloseSTT();
            return InitSTT(recognizerID);
        }

        public void Dispose()
        {
            CloseSTT();
        }

        #endregion

        #region Invocators

        private void OnWordRecognized(string word, float confidence)
        {
            var handler = WordRecognized;
            if (handler != null)
                handler(this, new SpeechDetectionEventArgs(word, confidence));
        }

        private void OnWordHypothesized(string word, float confidence)
        {
            var handler = WordHypothesized;
            if (handler != null)
                handler(this, new SpeechDetectionEventArgs(word, confidence));
        }

        private void OnSpeechDetected()
        {
            var handler = SpeechDetected;
            if (handler != null)
                handler(this, EventArgs.Empty);
        }

        private void OnRecognitionCompleted()
        {
            var handler = RecognitionCompleted;
            if (handler != null)
                handler(this, EventArgs.Empty);
        }

        #endregion

        public bool InitSTT(string recognizerID = null)
        {
            try
            {
                Initialized = false;

                if (string.IsNullOrEmpty(recognizerID))
                    _recognition = new SpeechRecognitionEngine();
                else
                    _recognition = new SpeechRecognitionEngine(recognizerID);

                _recognition.RecognizeCompleted += recognition_RecognizeCompleted;
                _recognition.SpeechDetected += recognition_SpeechDetected;
                _recognition.SpeechRecognized += recognition_SpeechRecognized;
                _recognition.SpeechHypothesized += recognition_SpeechHypothesized;

                _speechFormat = new SpeechAudioFormatInfo(_audioFormat.SampleRate, (AudioBitsPerSample)_audioFormat.BitRate, (AudioChannel)_audioFormat.Channels);

                // load grammar
                var voiceCommands = new Choices(_choices.ToArray());
                var grammarBuilder = new GrammarBuilder(voiceCommands);
                grammarBuilder.Culture = _recognition.RecognizerInfo.Culture;
                var choicesGrammar = new Grammar(grammarBuilder);

                _recognition.UnloadAllGrammars();
                _recognition.LoadGrammar(choicesGrammar);

                Initialized = true;
                return true;
            }
            catch (Exception ex)
            {
            }

            return false;
        }

        private void CloseSTT()
        {
            if (!Initialized)
                return;

            _recognition.RecognizeCompleted -= recognition_RecognizeCompleted;
            _recognition.SpeechDetected -= recognition_SpeechDetected;
            _recognition.SpeechRecognized -= recognition_SpeechRecognized;
            _recognition.SpeechHypothesized -= recognition_SpeechHypothesized;
            _recognition.Dispose();
            _recognition = null;

            Initialized = false;
        }

        void recognition_SpeechHypothesized(object sender, SpeechHypothesizedEventArgs e)
        {
            foreach (var word in e.Result.Words)
            {
                if (word.Confidence > 0.005f)
                    OnWordHypothesized(word.Text, word.Confidence);
            }
        }

        void recognition_SpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {
            foreach (var word in e.Result.Words)
            {
                if (word.Confidence > 0.25f)
                    OnWordRecognized(word.Text, word.Confidence);
            }
        }

        void recognition_SpeechDetected(object sender, SpeechDetectedEventArgs e)
        {
            OnSpeechDetected();
        }

        void recognition_RecognizeCompleted(object sender, RecognizeCompletedEventArgs e)
        {
            OnRecognitionCompleted();
        }

        private SpeechRecognizerInfo GetRecognizerInfo(RecognizerInfo recognizerInfo)
        {
            var info = new SpeechRecognizerInfo();
            info.ID = recognizerInfo.Id;
            info.Name = recognizerInfo.Name;
            info.Culture = recognizerInfo.Culture;
            foreach (var format in recognizerInfo.SupportedAudioFormats)
                info.SupportedAudioFormats.Add(new AudioFormat(format.SamplesPerSecond, format.ChannelCount, format.BitsPerSample));

            return info;
        }
    }
}
